/*! \file FG_battle.cpp

    \brief DynRPG plugin to implement battle system in The Legend of Zelda: Forgotten Gates

    \author Bernard J. Badger, AKA Aubrey the Bard

    FG_battle is a DynRPG plugin made specifically to implement the battle system for The Legend of
    Zelda: Forgotten Gates. It uses code from PepsiOtaku's ATB Overhaul plugin to control battler
    turns and avoid the skill override bug.

    TODO: target selection
    TODO: display status conditions above heroes and monsters
    TODO: display bars for monster HP, MP, and ATB above monsters
    TODO: reading battle data from external file, hopefully Excel spreadsheet
    TODO: skill implementation
    TODO: monster action-deciding system
    TODO: check whether it would be possible to change monsters without using transform
*/


#include <DynRPG/DynRPG.h>
#define NOT_MAIN_MODULE
#include <vector>
#include <sstream>

const int NUM_HEROES = 4;                               //!< Maximum number of heroes
const int NUM_MONSTERS = 8;                             //!< Maximum number of monsters
const int ATB_MAX = 300000;                             //!< Maximum value for an ATB bar before it's supposed to reset

std::map<std::string, std::string> configuration;       //!< Configuration data from the DynRPG.ini file

int confAddValue;                                       //!< Additional speed value for ATB bars
int confHeroSpeedVarM;                                  //!< Hero speed master variables reference
int confBattlerStartVar;                                //!<
int confMonsterSpeedVarM;
int confDefaultHeroSpeed;
int confDefaultMonsterSpeed;
int confAtbModeSwitch;
int confActiveSpeed;
int confWaitSpeed;
bool atbWait;
bool monsterAction = false;
bool longAction = false;
bool confAgilMult;
bool condCheckFail;
int confFreezeSwitch;
std::string condExceptionSet;

std::vector<int> exceptionVect;



unsigned char frameTimer = 0;

int heroATB[NUM_HEROES];
int monsterATB[NUM_MONSTERS];


bool onStartup( char *pluginName )
{

	configuration = RPG::loadConfiguration( pluginName );
	confAtbModeSwitch = atoi( configuration["ATBmodeSwitch"].c_str() );
	confAddValue = atoi( configuration["AddValue"].c_str() );
	confAgilMult = configuration["MultiplyAgility"] == "true";
    confHeroSpeedVarM = atoi( configuration["HeroSpeedMasterVar"].c_str() ); // Convert String to Int
    confMonsterSpeedVarM = atoi( configuration["MonsterSpeedMasterVar"].c_str() );
    confBattlerStartVar = atoi( configuration["BattlerStartVar"].c_str() );
    confDefaultHeroSpeed = atoi( configuration["DefaultHeroSpeed"].c_str() );
    confDefaultMonsterSpeed = atoi( configuration["DefaultMonsterSpeed"].c_str() );
    confActiveSpeed = atoi( configuration["ActiveSpeed"].c_str() );
    confWaitSpeed = atoi( configuration["WaitSpeed"].c_str() );
    confFreezeSwitch = atoi( configuration["FreezeSwitch"].c_str() );
    // get the condition exceptions
    condExceptionSet = configuration["ConditionExceptions"];
    std::string delimiter = ",";
    size_t pos = 0;
    int token;
    while( (pos = condExceptionSet.find( delimiter )) != std::string::npos )
    {

        token = atoi( condExceptionSet.substr( 0, pos ).c_str() );
        exceptionVect.push_back( token );
        condExceptionSet.erase( 0, pos + delimiter.length() );

    }
    token = atoi( condExceptionSet.c_str() );
    exceptionVect.push_back( token );
    // Initialize hero and monster ATB values
    for( int i = 0; i < NUM_HEROES; i++ )
    {

        heroATB[i] = 0;

    }
    for( int i = 0; i < NUM_MONSTERS; i++ )
    {

        monsterATB[i] = 0;

    }

	return true;

}

void initializeSpeedVars()
{

    if( RPG::variables[confHeroSpeedVarM] == 0 )
    {

        RPG::variables[confHeroSpeedVarM] = confDefaultHeroSpeed;

    }
    if( RPG::variables[confMonsterSpeedVarM] == 0 )
    {

        RPG::variables[confMonsterSpeedVarM] = confDefaultMonsterSpeed;

    }

}

void onNewGame()
{

    // initialize the master speed variables based on what's in DynRPG.ini
    initializeSpeedVars();

}

void onLoadGame( int id, char *data, int length )
{

    // initialize the master speed variables based on what's in DynRPG.ini
    initializeSpeedVars();
    // initialize the Active/Wait system based on what the ATB Mode Switch is in the save file
    if( RPG::switches[confAtbModeSwitch] == true )
    {

        RPG::system->atbMode = RPG::ATBM_ACTIVE;

    }
    else
    {

        RPG::system->atbMode = RPG::ATBM_WAIT;

    }

}

bool onSetSwitch( int id, bool value )
{

    // if ATBmodeSwitch is on, set it to active, otherwise if off, set to wait
    if( id == confAtbModeSwitch )
    { // this was moved from ragnadyn since it was more appropriate here

        if( value == true )
        {

            RPG::system->atbMode = RPG::ATBM_ACTIVE;

        }
        else
        {

            RPG::system->atbMode = RPG::ATBM_WAIT;

        }

    }
    return true;

}

// this will reset RPG::BattleSpeed. NOTE: RPG::battleSpeed is only set to 100 (default), 0 (for Wait Mode), or 75 (for Active Mode)
// onFrame only makes alterations to individual battler's ATB bars
void resumeAtb()
{

    if( atbWait == true )
    {

        RPG::battleSpeed = 100;
        atbWait = false;

    }

}

void onFrame( RPG::Scene scene )
{

    static bool heroActionReady;        // Indicator that a hero is ready to take an action

    if( scene == RPG::SCENE_BATTLE
        && !atbWait && !monsterAction && !longAction )
    {   // Do not do the following if the ATB bar is set to 0 or a monster is performing an action
        // or a hero is performing a long action

        frameTimer++;
        if( frameTimer > 10 )
        {

            frameTimer = 0;
            for( int i = 0; i < 8; i++ )
            {

                for( unsigned int j = 0; j < exceptionVect.size(); j++ )
                {

                    if( RPG::Actor::partyMember(i) != NULL
                        && RPG::Actor::partyMember(i)->conditions[exceptionVect[j]] != 0 )
                    {

                        RPG::Actor::partyMember(i)->atbValue = 0;
                        condCheckFail = true;
                        break;

                    }
                    else
                    {

                        condCheckFail = false;

                    }
                }
                // Heroes
                if( RPG::Actor::partyMember(i) != NULL && i < 4 && condCheckFail == false
                    // The ATB bar shouldn't move in these situations...
                    && ((RPG::Actor::partyMember(i)->conditions[1] == 0 // ... the hero is not dead
                    && RPG::Actor::partyMember(i)->actionStatus == RPG::AS_IDLE // ... the hero is idle
                    && RPG::Actor::partyMember(i)->animationId != 14) // ... the hero has won the battle
                    // 14 is the animation id for battle won for whatever reason-- should be 11
                    && RPG::switches[confFreezeSwitch] == false // Other plugins need a means of freezing battle
                    ) )
                {

                    if( RPG::Actor::partyMember(i)->atbValue < ATB_MAX )
                    {

                        if( confAgilMult == true )
                        {

                            RPG::Actor::partyMember( i )->atbValue = RPG::Actor::partyMember( i )->atbValue +
                                ((confAddValue * RPG::Actor::partyMember( i )->getAgility()) * RPG::variables[confBattlerStartVar + i]);

                        }
                        else
                        {

                            RPG::Actor::partyMember( i )->atbValue = RPG::Actor::partyMember( i )->atbValue +
                            (confAddValue * RPG::variables[confBattlerStartVar + i]);

                        }

                    }
                    else
                    {

                        RPG::Actor::partyMember( i )->atbValue = ATB_MAX;

                    }
                    if( RPG::Actor::partyMember( i )->atbValue < 0 )
                    {

                        RPG::Actor::partyMember( i )->atbValue = 0;

                    }

                }
                for(unsigned int j = 0; j < exceptionVect.size(); j++)
                {

                    if( RPG::monsters[i] != NULL
                        && RPG::monsters[i]->conditions[exceptionVect[j]] != 0 )
                    {

                        condCheckFail = true;
                        break;

                    }
                    else
                    {

                        condCheckFail = false;

                    }

                }
                // Monsters
                if( RPG::monsters[i] != NULL && condCheckFail == false
                    && ((RPG::monsters[i]->conditions[1] == 0 // the monster is dead
                    && RPG::monsters[i]->actionStatus == RPG::AS_IDLE)
                    && RPG::switches[confFreezeSwitch] == false // Other plugins need a means of freezing battle
                    ) )
                {

                    if( confAgilMult )
                    {

                        monsterATB[i] += ((confAddValue * RPG::monsters[i]->getAgility())
                            * RPG::variables[confBattlerStartVar + 4 + i]);

                    }
                    else
                    {

                        monsterATB[i] += (confAddValue * RPG::variables[confBattlerStartVar + 4 + i]);

                    }
                    if( monsterATB[i] < 0 )
                    {

                        monsterATB[i] = 0;

                    }
                    // Check for heroes being ready to act before allowing monster to act
                    heroActionReady = false;
                    for( int j = 0; j < NUM_HEROES; j++ )
                    {

                        if( RPG::Actor::partyMember( j )->atbValue >= ATB_MAX )
                        {   // Found a hero ready for an action

                            heroActionReady = true;

                        }

                    }
                    if( !heroActionReady )
                    {   // No heroes are ready to take an action; safe to let monsters take action

                        if( monsterATB[i] >= ATB_MAX )
                        {   // Monster is ready for action

                            // <DEBUG>
//                            MessageBox(
//                                    NULL, // We don't need a window handle
//                                    "Notice: monster ready for action.", // Text
//                                    "DEBUG", // Title
//                                    MB_OK // Flags
//                            );
                            // </DEBUG>
                            // Set monster's ATB value in RPG system to max
                            RPG::monsters[i]->atbValue = ATB_MAX;
                            // Decrease monster's ATB value in plugin by max
                            monsterATB[i] -= ATB_MAX;

                        }

                    }

                }

            }

        }

    }
    else if( scene != RPG::SCENE_BATTLE )
    {

        resumeAtb();

    }

}

// anytime the battle status window refreshes
bool onBattleStatusWindowDrawn( int x, int selection, bool selActive, bool isTargetSelection, bool isVisible )
{

    if( selActive == true )
    {

        if( RPG::system->atbMode == RPG::ATBM_WAIT )
        {

            RPG::battleSpeed = confWaitSpeed; // wait mode should stop the atb bar when heros are selecting actions

        }
        else
        {

            RPG::battleSpeed = confActiveSpeed; // active mode should still move the atb bar by the % set in confActiveSpeed (100 is the default speed)

        }
        atbWait = true;

    }

    // if the party member dies (or has a condition with a status restriction) while the status window was open
    for( unsigned int j = 0; j < exceptionVect.size(); j++ )
    {

        if( (RPG::Actor::partyMember( selection )->conditions[1]
            && RPG::Actor::partyMember( selection )->conditions[exceptionVect[j]]) != 0)
        {

            RPG::Actor::partyMember( selection )->atbValue = 0;
            resumeAtb(); // resume the ATB bar
            //MessageBox(NULL, "what happen", "Debug", MB_ICONINFORMATION);
            break;

        }

    }
    return true;

}
// resume the ATB bar when the action is started
bool onDoBattlerAction(RPG::Battler* battler, bool firstTry)
{

    if( firstTry == true )
    {

        // however, make an exception if a monster is starting an action
        if( battler->isMonster() == true )
        {

            monsterAction = true;

        }
        // or if the battler is doing a long action (added by AubreyTheBard)
        if( battler->action->kind == RPG::AK_SKILL || battler->action->kind == RPG::AK_ITEM )
        {

            RPG::battleSpeed = confWaitSpeed;
            longAction = true;     // Added by AubreyTheBard

        }

    }
    return true;

}

bool onBattlerActionDone(RPG::Battler* battler, bool success)
{

    if( success == true )
    {

        resumeAtb();
        monsterAction = false;
        longAction = false;

    }
    return true;

}

//! Respond to potential command overwrite cases
/*!
    onEventCommand() is called when the game runs an event command line. In this plugin, the
    function is used to prevent the actors' ATB bars from moving while event commands are running.

    \param scriptLine (RPG::EventScriptLine*) The event script line which is about to be executed
    \param scriptData (RPG::EventScriptData*) Pointer to the RPG::EventScriptData object of the current event script
    \param eventId (int) The ID of the current event (negative for common events, zero for battle events)
    \param pageId (int) The ID of the current event page (zero for common and battle events - sorry, no battle event page ID yet)
    \param lineId (int) The zero-based line number
    \param nextLineId (int*) Pointer to the next executed line number (-1 for default)
    \return (bool) false will prevent other plugins from receiving this notification, use true otherwise
*/
bool onEventCommand(RPG::EventScriptLine* scriptLine, RPG::EventScriptData* scriptData,
                    int eventId, int pageId, int lineId, int* nextLineId)
{

    if(eventId == 0)
    { // battle events

        if( scriptData->currentLineId == 0 )
        {

            atbWait = true; // when an event page starts

        }
        else if( scriptLine->command == RPG::EVCMD_END_OF_EVENT )
        { // when an event page ends

            resumeAtb();

        }

    }
    return true;

}

//! Respond to potential comment commands
/*!
    onComment() is called when the game runs across a comment line in its event scripting. Comment
    commands are how the user communicates with the plugin, so this function checks for comment
    commands and responds accordingly.

    \param text (const char*) The comment's content as simple text
    \param parsedData (const RPG::ParsedCommentData*) The already parsed data
    \param nextScriptLine (RPG::EventScriptLine*) The next event script line after the comment
    \param scriptData (RPG::EventScriptData*) Pointer to the RPG::EventScriptData object of the current event script
    \param eventId (int) The ID of the current event (negative for common events, zero for battle events)
    \param pageId (int) The ID of the current event page (zero for common and battle events - sorry, no battle event page ID yet)
    \param lineId (int) The zero-based line number
    \param nextLineId (int*) Pointer to the next executed line number (-1 for default)
    \return (bool) false will prevent other plugins from receiving this notification, use true otherwise
*/
bool onComment ( const char* text,
                 const RPG::ParsedCommentData* parsedData,
                 RPG::EventScriptLine* nextScriptLine,
                 RPG::EventScriptData* scriptData,
                 int eventId,
                 int pageId,
                 int lineId,
                 int* nextLineId )
{

    std::string cmd = parsedData->command;
    if( !cmd.compare("init_hero_speed") )
    {

        for(int i=0; i<4; i++)
        {

            RPG::variables[confBattlerStartVar+i] = RPG::variables[confHeroSpeedVarM];

        }

    }
    if(!cmd.compare("init_monster_speed"))
    {

        for(int i=0; i<8; i++)
        {

            RPG::variables[confBattlerStartVar+4+i] = RPG::variables[confMonsterSpeedVarM];

        }

    }
    if( !cmd.compare("condition_speed_check") )
    {

        // parameter 1: condition ID
        if( parsedData->parametersCount >= 3 && parsedData->parameters[0].type == RPG::PARAM_NUMBER
            && parsedData->parameters[1].type == RPG::PARAM_NUMBER
            && parsedData->parameters[2].type == RPG::PARAM_NUMBER )
        {

            for( int i = 0; i < 8; i++ )
            {

                if( i < 4 )
                {

                    if( RPG::Actor::partyMember(i) != NULL
                        && RPG::Actor::partyMember(i)->conditions[parsedData->parameters[0].number] != 0 )
                    {

                        RPG::variables[confBattlerStartVar+i] = parsedData->parameters[1].number;

                    }

                }
                if( RPG::monsters[i] != NULL
                    && RPG::monsters[i]->conditions[parsedData->parameters[0].number] != 0 )
                {

                    RPG::variables[confBattlerStartVar+4+i] = parsedData->parameters[2].number;

                }

            }

        }

    }
    return true;

}

//! Clean up after use
/*!
    onExit() is called when the game closes. In this plugin this is used to perform any needed
    cleanup.
*/
void onExit()
{

    exceptionVect.clear();

}
